%% $Id: SetOps.oz,v 1.1 2007/10/20 23:58:29 leavens Exp leavens $
%% This functor should be compiled with the command:
%%
%%    ozc -c SetOps.oz
%%
%% Or use the "Compile file" command from the Oz menu in the Mozart IDE.
%% Don't "feed" the buffer or file to Oz, as that will cause Oz to complain
%% about a "syntax error" in the first line below.
functor $
export
   emptySet: EmptySet
   asSet: AsSet
   isEmpty: IsEmpty
   size: Size
   choose: Choose
   isMember: IsMember
   isSubset: IsSubset
   equal: Equal
   add: Add
   remove: Remove
   union: Union
   minus: Minus
   intersect: Intersect
   unionList: UnionList
define
   %%-------------------------------------------------------
   %% Do not change any of the following provided functions.
   %% The code you are to write follows them   
   %%-------------------------------------------------------

   %% We are representing sets as lists, without duplicates

   %% Return the empty set
   fun {EmptySet} nil end

   %% Return a set containing exactly the distinct elements of L
   fun {AsSet L}
      case L of
         E|Es then {Add {AsSet Es} E}
      else nil
      end
   end

   %% Is this set empty?
   fun {IsEmpty L}
      case L of
         _|_ then false
      else true
      end
   end

   %% Return the number of distinct elements in this set.
   fun {Size L}
      case L of
         _|Es then 1 + {Size Es}
      else 0
      end
   end

   %% Return an element of the set
   fun {Choose L}
      %% requires: L is not empty
      case L of
         E|_ then E
      else raise cantChooseFromEmpty end
      end
   end

   %% Is X a member of L
   fun {IsMember L X}
      case L of
         E|Es then X == E orelse {IsMember Es X}
      else false
      end
   end

   %% Is S1 a (not necessarily proper) subset of S2?
   %% That is, are all elements of S1 also elements of S2?
   fun {IsSubset S1 S2}
      case S1 of
         E|Es then {IsMember S2 E} andthen {IsSubset Es S2}
      else true
      end
   end

   %% Are sets S1 and S2 equal?
   fun {Equal S1 S2}
      {Size S1} == {Size S2} andthen {IsSubset S1 S2}
   end

   %%----------------------------------
   %% End of provided code
   %%----------------------------------

   %%---------------------------------
   %% Fill in your code for each of the functions below
   %%---------------------------------

   %% Return a new set containing just X and all the elements of S
   fun {Add S X}
      if {IsMember S X}
      then S
      else X|S
      end
   end

   %% Return a new set that contains all the elements of S except X
   fun {Remove S X}
      case S of
         E|Es then if X==E
                   then Es
                   else E|{Remove Es X}
                   end
      else nil
      end
   end

   %% Return a set that contains just the elements of both S1 and S2
   fun {Union S1 S2}
      case S1 of
         E|Es then {Union Es {Add S2 E}}
      else S2
      end
   end

   %% Return a set that contains just the elements of S1 that are not in S2
   fun {Minus S1 S2}
      case S1 of
         E|Es then if {IsMember S2 E} then {Minus Es S2}
                   else {Add {Minus Es S2} E}
                   end
      else nil
      end
   end

   %% Return a set that contains just the elements of S1 that are also in S2
   fun {Intersect S1 S2}
      case S1 of
         E|Es then if {IsMember S2 E}
                   then {Add {Intersect Es S2} E}
                   else {Intersect Es S2}
                   end
      else nil
      end
   end

   %% Return a set that contains just the elements of all the sets in Sets
   fun {UnionList Sets}
      case Sets of
         S|Ss then {Union S {UnionList Ss}}
      else nil
      end
   end
end
